#include <FK/FK.h>
#include "TrackBall.h"

// n  min ȏ max ȉɂ܂߂
double Clamp(double n, double min, double max)
{
	if(n < min) return min;
	if(n > max) return max;
	return n;
}

//  P1Q1  P2Q2 ɂŋߐړ_ C1  C2 ߂
// P1Q1  C1 ̈ʒu s ƁAP2Q2  C2 ̈ʒu t Ԃ
// Ԃl C1  C2 ̋̕ƂȂ
double ClosestPtSegmentSegment(fk_Vector p1, fk_Vector q1, fk_Vector p2, fk_Vector q2,
							   double &s, double &t, fk_Vector &c1, fk_Vector &c2)
{
	fk_Vector d1 = q1 - p1;
	fk_Vector d2 = q2 - p2;
	fk_Vector r = p1 - p2;
	double a = d1*d1;
	double e = d2*d2;
	double f = d2*r;
	if(a < FK_EPS && e < FK_EPS) {
		s = t = 0.0;
		c1 = p1;
		c2 = p2;
	}
	if(a < FK_EPS) {
		s = 0.0;
		t = f/e;
		t = Clamp(t, 0.0, 1.0);
	} else {
		double c = d1*r;
		if(e < FK_EPS) {
			t = 0.0;
			s = Clamp(-c/a, 0.0, 1.0);
		} else {
			double b = d1*d2;
			double denom = a*e - b*b;
			if(denom != 0.0) {
				s = Clamp((b*f - c*e)/denom, 0.0, 1.0);
			} else {
				s = 0.0;
			}
			t = (b*s + f)/e;
			if(t < 0.0) {
				t = 0.0;
				s = Clamp(-c/a, 0.0, 1.0);
			} else if(t > 1.0) {
				t = 1.0;
				s = Clamp((b - c)/a, 0.0, 1.0);
			}
		}
	}
	c1 = p1 + d1*s;
	c2 = p2 + d2*t;
	return (c1 - c2)*(c1 - c2);
}

class CCapsule {
	fk_Sphere	spr;
	fk_Prism	pri;
	fk_Model	visModel[3];
public:
	fk_Vector	p, q;
	double		r, h;

	CCapsule(void)
	{
		spr.setDivide(4);
		pri.setDivide(8);
		visModel[0].setShape(&spr);
		visModel[1].setShape(&spr);
		visModel[2].setShape(&pri);
	};

	void show(fk_Scene *argScene)
	{
		argScene->entryModel(&visModel[2]);
		argScene->entryModel(&visModel[1]);
		argScene->entryModel(&visModel[0]);
	};

	void hide(fk_Scene *argScene)
	{
		argScene->removeModel(&visModel[2]);
		argScene->removeModel(&visModel[1]);
		argScene->removeModel(&visModel[0]);
	};

	void update(void)
	{
		spr.setRadius(r);
		pri.setTopRadius(r);
		pri.setBottomRadius(r);
		pri.setHeight((q-p).dist());
		visModel[0].glMoveTo(p);
		visModel[1].glMoveTo(q);
		visModel[2].glMoveTo(p);
		visModel[2].glVec(q-p);
	};

	bool isContact(CCapsule &other, fk_Vector *revVec)
	{
		static fk_Vector	c1, c2;
		static double		s, t, retVal, tmpDist;

		retVal = ClosestPtSegmentSegment(p, q, other.p, other.q, s, t, c1, c2);

		if(revVec != NULL) {
			*revVec = c2-c1;
			tmpDist = (*revVec).dist() - (r + other.r);
			(*revVec).normalize();
			*revVec *= tmpDist;
		}

		return bool(retVal < pow(r + other.r, 2.0));
	};
};

int main(int argc, char *argv[])
{
	Fl_Window		mainWindow(512, 512, "FK TEST");
	fk_Model		camera, blockModel, lightModel;
	fk_Light		light;
	fk_Block		block(500.0, 1.0, 500.0);
	fk_Scene		scene;
	fk_Window		window(0, 0, 512, 512);

	fk_TrackBall	tb(&window, &camera);

	CCapsule		capA, capB;
	fk_Vector		revVec;

	mainWindow.end();
	fk_InitMaterial();

	// Ɩ̐ݒ
	lightModel.setShape(&light);
	lightModel.setMaterial(White);
	lightModel.glMoveTo(0.0, 0.0, 0.0);
	lightModel.glFocus(-1.0, -1.0, -1.0);

	// ̂̐ݒ
	blockModel.setShape(&block);
	blockModel.setMaterial(Yellow);

	// effBXvCXgɓo^
	scene.entryCamera(&camera);
	scene.entryModel(&blockModel);
	scene.entryModel(&lightModel);

	// EBhEփfBXvCXgo^
	window.setScene(&scene);

	// _̈ʒuƎpݒ
	camera.glMoveTo(0.0, 0.0, 400.0);
	camera.glFocus(0.0, 0.0, 0.0);
	camera.glUpvec(0.0, 1.0, 0.0);

	capA.p.set(-100.0, 0.0, 0.0);
	capA.q.set(-100.0, 50.0, 0.0);
	capA.r = 20.0;
	capB.p.set(100.0, 0.0, 0.0);
	capB.q.set(100.0, 0.0, 50.0);
	capB.r = 10.0;

	capA.show(&scene);
	capB.show(&scene);

	mainWindow.show();
	window.show();

	while(true) {

		if(mainWindow.visible() == 0) {
			if(Fl::wait() == 0) {
				break;
			} else {
				continue;
			}
		}

		if(window.drawWindow() == 0) break;
		if(Fl::check() == 0) break;
		if(window.winOpenStatus() == false) continue;

		tb.update();

		capA.update();
		capB.update();

		if(window.getSpecialKeyStatus(FK_UP)) {
			capB.p.z -= 1.0;
			capB.q.z -= 1.0;
		}
		if(window.getSpecialKeyStatus(FK_DOWN)) {
			capB.p.z += 1.0;
			capB.q.z += 1.0;
		}
		if(window.getSpecialKeyStatus(FK_LEFT)) {
			capB.p.x -= 1.0;
			capB.q.x -= 1.0;
		}
		if(window.getSpecialKeyStatus(FK_RIGHT)) {
			capB.p.x += 1.0;
			capB.q.x += 1.0;
		}

		if(capA.isContact(capB, &revVec)) {
			cout << "Hit:" << revVec.dist() << endl;
			capB.p -= revVec;
			capB.q -= revVec;
			capB.update();
		}
	}

	return 0;
}
